#!/usr/bin/env python
#
# Copyright (c) 2009 Heikki Toivonen <my first name at heikkitoivonen.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['SecretError', 'NoDataError', 'DecryptionError', 'EncryptionError',
           'PasswordError', 'Secret']

"""
Small utility and module for encrypting and decrypting data using symmetric-key
algorithms. By default uses 256-bit AES (Rijndael) using CBC, but some options
are configurable. PBKDF2 algorithm used to derive key from password.

Sample uses: passwords in INI files, password manager, encrypted files
"""

import os
import sys
import hmac
import string

import hashlib
from binascii import hexlify, unhexlify

import six
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import algorithms, modes
from cryptography.hazmat.primitives.ciphers.base import Cipher

CRYPTO_BACKEND = default_backend()


# partially copied from Django
def force_bytes(s, encoding='utf-8', errors='strict'):
    if isinstance(s, bytes):
        if encoding == 'utf-8':
            return s
        else:
            return s.decode('utf-8', errors).encode(encoding, errors)
    if not isinstance(s, six.string_types):
        try:
            if six.PY3:
                return six.text_type(s).encode(encoding)
            else:
                return bytes(s)
        except UnicodeEncodeError:
            if isinstance(s, Exception):
                return b' '.join(force_bytes(arg, encoding, strings_only, errors)
                                 for arg in s)
            return six.text_type(s).encode(encoding, errors)
    else:
        return s.encode(encoding, errors)


class SecretError(Exception):
    """
    Base class for Secret specific errors.
    """

class NoDataError(SecretError):
    """
    This exception will be raised if we don't have data to encrypt/decrypt.
    """

class DecryptionError(SecretError):
    """
    Failed to decrypt. Can happen with wrong password, for example.
    """

class EncryptionError(SecretError):
    """
    Failed to encrypt.
    """

class PasswordError(SecretError):
    """
    Problem with password(s).
    """


class Secret:
    """
    Stores a secret and has ways to decrypt the secret and set new secret.

    @warning: Once password is in memory, it will be possible
              to get the ciphertext as well. This may be possible
              even after the password is cleared due to Python memory
              management.
    @warning: If the password is used, secret will be decrypted and available
              in plain text in memory, possibly even after it has been
              explicitly cleared after use.
    @warning: If weak password is used, the
              encryption will not be of much help.
    """
    def __init__(self, iv=None, salt=None, ciphertext=None,
                 iterations=1000):
        """
        Construct a Secret object.

        ciphertext, iv and salt can be None when originally created. The Secret
        is then considered to not hold any data. To set new data, call
        encrypt(). It'll use AES 256 CBC as the algorithm.

        @param iv: The IV, 256 bits (byte string 32 long)
        @param salt: The salt, 256 bits (byte string 32 long)
        @param ciphertext: The secret to hold
        @param iterations: The number of iterations to use with PBKDF2,
                           recommend 1000.
        @param param: The algorithm to use, recommend .
        """
        self.ciphertext = ciphertext
        self.iv = iv or os.urandom(32)
        self.salt = salt or os.urandom(32)
        self.iterations = iterations

    def decrypt(self, password):
        """
        Decrypt.

        @param password: The password to decrypt data with.
        @return: Decrypted data
        """
        if not self.ciphertext or not self.iv or not self.salt or \
            password is None:
            raise NoDataError

        # If the password is callable we'll assume it will return the
        # real password.
        try:
            password = password()
        except TypeError:
            pass

        # the crypto algorithms are unicode unfriendly
        password = force_bytes(password)

        # derive 256 bit key using the pbkdf2 standard
        key = hashlib.pbkdf2_hmac(
            'sha1', password, self.salt, self.iterations, 32)

        # Derive encryption key and HMAC key from it
        # See Practical Cryptography section 8.4.1.
        hmacKey = hashlib.sha256(key + b'MAC').digest()
        encKey = hashlib.sha256(key + b'encrypt').digest()
        del key

        # decrypt
        ret = decrypt(self.ciphertext, encKey, self.iv)
        del encKey

        # Check MAC
        mac = ret[-64:]
        ret = ret[:-64]
        try:
            invalid_mac = force_bytes(hmac.new(
                hmacKey,
                ret + self.iv + self.salt,
                hashlib.sha256
            ).hexdigest()) != mac

            if invalid_mac:
                raise DecryptionError('HMAC does not match')
        finally:
            del hmacKey

        return ret

    def encrypt(self, cleartext, password):
        """
        Encrypt.

        @param cleartext: The data to encrypt.
        @param password: The password to encrypt data with.
        @return: Encrypted data
        """
        if cleartext is None or password is None:
            raise NoDataError

        # If the password is callable we'll assume it will return the
        # real password.
        try:
            password = password()
        except TypeError:
            pass

        # the crypto algorithms are unicode unfriendly
        password = force_bytes(password)
        cleartext = force_bytes(cleartext)

        # get 256 bit random encryption salt
        self.salt = os.urandom(32)

        # derive 256 bit encryption key using the pbkdf2 standard
        key = hashlib.pbkdf2_hmac(
            'sha1', password, self.salt, self.iterations, 32)

        # Derive encryption key and HMAC key from it
        # See Practical Cryptography section 8.4.1.
        hmacKey = hashlib.sha256(key + b'MAC').digest()
        encKey = hashlib.sha256(key + b'encrypt').digest()
        del key

        # get 256 bit random iv
        self.iv = os.urandom(32)

        # Add HMAC to cleartext so that we can check during decrypt if we got
        # the right cleartext back. We are doing sign-then-encrypt, which let's
        # us encrypt empty cleartext (otherwise we'd need to pad with some
        # string to encrypt). Practical Cryptography by Schneier & Ferguson
        # also recommends doing it in this order in section 8.2.
        mac = force_bytes(hmac.new(
            hmacKey,
            cleartext + self.iv + self.salt,
            hashlib.sha256
        ).hexdigest())

        del hmacKey

        # encrypt
        self.ciphertext = encrypt(cleartext + mac, encKey, self.iv)

        return self.ciphertext

    def serialize(self, serialize=None):
        """Serialize secret.

        @param serialize: None or callable that must accept string to serialize
        @return: Serialized string
        """
        if not self.ciphertext or not self.iv or not self.salt:
            raise NoDataError

        serialized = b'%s|%s|%s' % (hexlify(self.iv), hexlify(self.salt),
                                    hexlify(self.ciphertext))
        if serialize is not None:
            serialize(serialized)
        return serialized

    def deserialize(self, deserialize):
        """Deserialize secret.

        @param deserialize: String or callable that must return the serialized form.
        """
        try:
            serialized = deserialize()
        except TypeError:
            serialized = deserialize

        iv, salt, ciphertext = force_bytes(serialized).split(b'|')
        self.iv, self.salt, self.ciphertext = unhexlify(iv), unhexlify(salt), unhexlify(ciphertext)

    def clear(self):
        try:
            del self.ciphertext
        except AttributeError:
            pass
        try:
            del self.iv
        except AttributeError:
            pass
        try:
            del self.salt
        except AttributeError:
            pass


def decrypt(ciphertext, key, iv):
    """
    Decrypt ciphertext
    """
    assert len(key) == len(iv) == 32

    cipher = Cipher(
        algorithms.AES(key), modes.CBC(iv[:16]), CRYPTO_BACKEND).decryptor()
    del key
    plaintext_padded = cipher.update(ciphertext) + cipher.finalize()

    unpadder = padding.PKCS7(algorithms.AES.block_size).unpadder()
    plaintext = unpadder.update(plaintext_padded) + unpadder.finalize()

    return plaintext


def encrypt(plaintext, key, iv):
    """
    Encrypt plaintext
    """
    assert len(key) == len(iv) == 32

    cipher = Cipher(
        algorithms.AES(key), modes.CBC(iv[:16]), CRYPTO_BACKEND).encryptor()
    del key

    padder = padding.PKCS7(algorithms.AES.block_size).padder()
    padded_data = padder.update(plaintext) + padder.finalize()

    ciphertext = cipher.update(padded_data) + cipher.finalize()

    assert ciphertext
    return ciphertext


def get_password(confirm=True):
    """get password"""
    import getpass
    password = getpass.getpass('password:')
    if confirm:
        if getpass.getpass('password (again):') != password:
            raise PasswordError('Passwords do not match')
    return password


def main():
    from optparse import OptionParser

    usage = 'usage: %prog [options]'
    parser = OptionParser(usage=usage,
                          version='%prog ' + version,
                          description='Encrypt or decrypt data with password using 256 bit AES (Rijndael) encryption in CBC mode. Key derived from password with PBKDF2 algorithm using 1000 iterations.')
    parser.add_option('-d', '--decrypt',
                      action='store_true', dest='decrypt', default=False,
                      help="Decryption mode.")
    parser.add_option('-e', '--encrypt',
                      action='store_true', dest='encrypt', default=False,
                      help="Encryption mode.")
    parser.add_option('-i', '--in',
                      dest='infile', metavar='INFILE',
                      help='INFILE to read in. Without this options reads stdin.')
    parser.add_option('-o', '--out',
                      dest='outfile', metavar='OUTFILE',
                      help="OUTFILE to output into. Without this option prints to stdout.")
    parser.add_option('-p', '--password',
                      metavar='PASSWORD', dest='password',
                      help="Supply PASSWORD from the command line (otherwise will be prompted for). Try to not use this option, since it is safer to be prompted for password.")

    (options, args) = parser.parse_args()
    if (options.decrypt and options.encrypt) or (not options.decrypt and not options.encrypt):
        parser.print_help()
        sys.exit(1)

    if not options.infile:
        options.args = ''.join(sys.stdin.readlines())
    else:
        options.args = ''

    if options.encrypt:
        secret = Secret()
        secret.encrypt(options.args or open(options.infile, 'rb').read(),
                       options.password or get_password)
        if options.outfile:
            f = open(options.outfile, 'wb')
            secret.serialize(f.write)
            f.close()
        else:
            secret.serialize(sys.stdout.write)
    else:
        secret = Secret()
        secret.deserialize((options.args or open(options.infile, 'rb').read()).strip())
        cleartext = secret.decrypt(options.password or (lambda: get_password(confirm=False)))
        if options.outfile:
            f = open(options.outfile, 'wb')
            f.write(cleartext)
            f.close()
        else:
            sys.stdout.write(cleartext)


if __name__ == "__main__":
    main()



import os
import unittest

from binascii import hexlify

from m2secret import Secret

import pytest
from mock import patch


class TestSecret:
    """Verify the behavior of Secret class."""

    @pytest.fixture(autouse=True)
    def setup(self):
        self.iv = 'iv_45678123456781234567812345678'
        self.salt = 'salt____123456781234567812345678'
        self.ciphertext = 'yadda yadda'
        self.iterations = 1000
        assert len(self.iv) == 32
        assert len(self.salt) == 32

    def test_ascii(self):
        secret = Secret()
        secret.encrypt('foo', 'some-key')
        res = secret.serialize()

        reverse = Secret()
        reverse.deserialize(res)
        assert reverse.decrypt('some-key') == b'foo'

    def test_unicode(self):
        secret = Secret()
        secret.encrypt(u'foo', 'some-key')
        res = secret.serialize()

        reverse = Secret()
        reverse.deserialize(res)
        assert reverse.decrypt('some-key') == b'foo'

    def test_init(self):
        secret = Secret(
            iv=self.iv,
            salt=self.salt,
            ciphertext=self.ciphertext,
            iterations=self.iterations)

        assert secret.salt == self.salt
        assert secret.iv == self.iv
        assert secret.ciphertext == self.ciphertext
        assert secret.iterations == self.iterations

    @patch.object(os, 'urandom')
    def test_default_params(self, urandom):
        """Known defaults should be used when no parameters are used."""
        urandom.return_value = 'x' * 32
        secret = Secret()
        assert len(secret.iv) == 32
        assert len(secret.salt) == 32
        urandom.assert_called_with(32)
        assert secret.ciphertext is None
        assert secret.iterations == 1000

    @pytest.mark.parametrize(('expected', 'serialized_secret'), [
        # The `serialized_secret` values were generated on different machines
        # with different openssl versions and with the old m2crypto
        # version to make sure we don't break anything.
        (
            b'super secret password string from machine 1',
            '3f3a0bd330b9b64456e7a01bf0c767268476dff84c942c58ad899c7e37c72a4c|'
            '6846b0fce9f9ae3a066f8c5ca7deb9695a80902cefdcd4e06225e2f45a2bae32|'
            'dfb7e0f57395c42e3c3520c437669430bf75d3ac728010a3c8c8e1866c73a1eeb'
            '899bec27610688c5c9557c975bbdd4d89fb931de7eeaffb80d9542e0761347dd7'
            '8d2584ad6ad2b1eb883ccb1de3ab96092fc99ba43633ab05ce2bc8f99fbe377fe'
            '843f35ac5710716dbf4a110e2fc3b'
        ),
        (
            b'super secret password string from machine 1',
            '9b27af9136b04bf4ef6a0afeb0b402925b191882f93f3fd98c12eda214e684a7|'
            '8926f70d957753d543a1e0e5c6c2da1cf2eb92d51e026fdcc971757e58e7d7e3|'
            'f7707712f1ace51733581aa638a3f612fdaab60284610fe5b3bae4df06230be71'
            '090b308581734000a3c093597288cdf7d844829932478962550ebbcb903c9e0d9'
            '0d8346d1c364f805295ef2cc669936b6ce9745188f92f4a89813b0bf0a98cbf3b'
            '9e1e60c36e841e3ac01c221f938d6'
        ),
    ])
    def test_legacy_compatibility(self, expected, serialized_secret):
        secret = Secret()
        secret.deserialize(serialized_secret)
        assert secret.decrypt('super-secret-encryption-key') == expected
